from pathlib import Path
from typing import Any

import deltalake
import numpy as np
import pandas as pd
import pytest
import yaml


@pytest.fixture()
def delta_table_data() -> pd.DataFrame:
    periods = 24
    daterange = pd.date_range(
        start="2025-01-01T00:00:00+00:00", periods=periods, freq="1h"
    )
    temperature = np.round(np.linspace(-5, -5 + periods, num=periods))
    return pd.DataFrame(
        {"time": daterange, "city": "Grenoble", "temperature": temperature}
    )


@pytest.fixture()
def delta_table(tmp_path: Path, delta_table_data: pd.DataFrame) -> deltalake.DeltaTable:
    table_path = tmp_path / "delta_table"
    schema = deltalake.Schema(
        [
            deltalake.Field("time", "timestamp_ntz", nullable=False),  # type: ignore[arg-type]
            deltalake.Field("city", "string", nullable=False),  # type: ignore[arg-type]
            deltalake.Field("temperature", "float", nullable=False),  # type: ignore[arg-type]
        ]
    )
    dt = deltalake.DeltaTable.create(
        table_path, schema, name="delta_table", description="Sample Delta Table"
    )
    deltalake.write_deltalake(dt, delta_table_data, mode="append")
    return dt


@pytest.fixture()
def sample_config(delta_table: deltalake.DeltaTable) -> dict[str, Any]:
    return {
        "tables": [
            {"name": "delta_table", "uri": delta_table.table_uri, "format": "delta"},
            {
                "name": "123_delta_table",  # test table name starting with digits
                "uri": delta_table.table_uri,
                "format": "delta",
            },
            {
                "name": "invalid_uri_table",
                "uri": "path/to/invalid_uri_table",
                "format": "delta",
            },
        ],
        "queries": [
            {
                "name": "daily_average_temperature",
                "title": "Daily average temperature",
                "sql": "select date_trunc('day', time) as day, round(avg(temperature)) as avg_temperature from delta_table group by day order by day asc",
            },
            {
                "name": "daily_average_temperature_params",
                "title": "Daily average temperature with parameters",
                "description": "Display daily average temperature values from `weather` table, with dynamic filters for start and end dates",
                "parameters": {
                    "start_date": {"default": "2025-01-01"},
                    "end_date": {"default": "2025-01-31"},
                },
                "sql": "select date_trunc('day', time) as day, round(avg(temperature)) as avg_temperature from delta_table where day between $start_date and $end_date group by day order by day asc",
            },
        ],
    }


@pytest.fixture()
def sample_config_table_delta_s3() -> dict[str, Any]:
    return {
        "name": "delta_table_s3",
        "uri": "s3://bucket/path/to/table",
        "format": "delta",
        "connection": {
            "s3": {
                "s3_access_key_id": "s3-access-key-id",
                "s3_secret_access_key": "s3-secret-access-key",
                "s3_region": "s3-region",
                "s3_endpoint_url": "https://s3.domain.com/",
                "s3_allow_http": False,
            },
        },
    }


@pytest.fixture()
def sample_config_table_delta_adls() -> dict[str, Any]:
    return {
        "name": "delta_table_adls",
        "uri": "abfss://container/path/to/table",
        "format": "delta",
        "connection": {
            "adls": {
                "adls_account_name": "adls-account-name",
                "adls_access_key": "adls-access-key",
                "adls_sas_key": "adls-sas-key",
                "adls_tenant_id": "adls-tenant-id",
                "adls_client_id": "adls-client-id",
                "adls_client_secret": "adls-client-secret",
                "azure_msi_endpoint": "https://msi.azure.com/",
                "use_azure_cli": False,
            },
        },
    }


@pytest.fixture()
def sample_config_path(tmp_path: Path, sample_config: dict[str, Any]) -> Path:
    config_path = tmp_path / "laketower.yml"
    config_path.write_text(yaml.dump(sample_config))
    return config_path



from pathlib import Path
from typing import Any

import pydantic
import pytest
import yaml

from laketower import config


def test_load_yaml_config(
    sample_config: dict[str, Any], sample_config_path: Path
) -> None:
    conf = config.load_yaml_config(sample_config_path)

    for table, expected_table in zip(conf.tables, sample_config["tables"], strict=True):
        assert table.name == expected_table["name"]
        assert table.uri == expected_table["uri"]
        assert table.table_format.value == expected_table["format"]
        assert table.connection is None

    for query, expected_query in zip(
        conf.queries, sample_config["queries"], strict=True
    ):
        assert query.name == expected_query["name"]
        assert query.title == expected_query["title"]
        assert query.description == expected_query.get("description")
        assert query.sql == expected_query["sql"]


def test_load_yaml_config_table_delta_s3(
    tmp_path: Path,
    sample_config: dict[str, Any],
    sample_config_table_delta_s3: dict[str, Any],
) -> None:
    sample_config["tables"] = [sample_config_table_delta_s3]
    sample_config_path = tmp_path / "laketower.yml"
    sample_config_path.write_text(yaml.dump(sample_config))

    conf = config.load_yaml_config(sample_config_path)
    assert len(conf.tables) == 1

    table = conf.tables[0]
    assert table.name == sample_config_table_delta_s3["name"]
    assert table.uri == sample_config_table_delta_s3["uri"]
    assert table.table_format.value == sample_config_table_delta_s3["format"]
    assert table.connection is not None
    assert table.connection.s3 is not None

    expected_s3_conn = sample_config_table_delta_s3["connection"]["s3"]
    table_s3_conn = table.connection.s3
    assert table_s3_conn.s3_access_key_id == expected_s3_conn["s3_access_key_id"]
    assert (
        table_s3_conn.s3_secret_access_key.get_secret_value()
        == expected_s3_conn["s3_secret_access_key"]
    )
    assert table_s3_conn.s3_region == expected_s3_conn["s3_region"]
    assert str(table_s3_conn.s3_endpoint_url) == expected_s3_conn["s3_endpoint_url"]
    assert table_s3_conn.s3_allow_http == expected_s3_conn["s3_allow_http"]


def test_load_yaml_config_table_delta_adls(
    tmp_path: Path,
    sample_config: dict[str, Any],
    sample_config_table_delta_adls: dict[str, Any],
) -> None:
    sample_config["tables"] = [sample_config_table_delta_adls]
    sample_config_path = tmp_path / "laketower.yml"
    sample_config_path.write_text(yaml.dump(sample_config))

    conf = config.load_yaml_config(sample_config_path)
    assert len(conf.tables) == 1

    table = conf.tables[0]
    assert table.name == sample_config_table_delta_adls["name"]
    assert table.uri == sample_config_table_delta_adls["uri"]
    assert table.table_format.value == sample_config_table_delta_adls["format"]
    assert table.connection is not None
    assert table.connection.adls is not None

    expected_adls_conn = sample_config_table_delta_adls["connection"]["adls"]
    table_adls_conn = table.connection.adls
    assert table_adls_conn.adls_account_name == expected_adls_conn["adls_account_name"]
    assert table_adls_conn.adls_access_key and (
        table_adls_conn.adls_access_key.get_secret_value()
        == expected_adls_conn["adls_access_key"]
    )
    assert table_adls_conn.adls_sas_key and (
        table_adls_conn.adls_sas_key.get_secret_value()
        == expected_adls_conn["adls_sas_key"]
    )
    assert table_adls_conn.adls_tenant_id == expected_adls_conn["adls_tenant_id"]
    assert table_adls_conn.adls_client_id == expected_adls_conn["adls_client_id"]
    assert table_adls_conn.adls_client_secret and (
        table_adls_conn.adls_client_secret.get_secret_value()
        == expected_adls_conn["adls_client_secret"]
    )
    assert (
        str(table_adls_conn.azure_msi_endpoint)
        == expected_adls_conn["azure_msi_endpoint"]
    )
    assert table_adls_conn.use_azure_cli == expected_adls_conn["use_azure_cli"]


def test_load_yaml_config_table_connection_mutually_exclusive(
    tmp_path: Path,
    sample_config: dict[str, Any],
    sample_config_table_delta_s3: dict[str, Any],
    sample_config_table_delta_adls: dict[str, Any],
) -> None:
    sample_config_table_delta_remote = {
        "name": "remote_delta_table",
        "uri": "somewhere",
        "format": "delta",
        "connection": sample_config_table_delta_s3["connection"]
        | sample_config_table_delta_adls["connection"],
    }
    sample_config["tables"] = [sample_config_table_delta_remote]
    sample_config_path = tmp_path / "laketower.yml"
    sample_config_path.write_text(yaml.dump(sample_config))

    with pytest.raises(
        pydantic.ValidationError,
        match="only one connection type can be specified among: 's3', 'adls'",
    ):
        config.load_yaml_config(sample_config_path)


@pytest.mark.parametrize(
    ("var_name", "var_value", "expected_result"),
    [
        ("TEST_STRING", "test_value", "test_value"),
        ("TEST_JSON", '{"key": "value", "number": 42}', {"key": "value", "number": 42}),
        ("TEST_ARRAY", '["item1", "item2", 42]', ["item1", "item2", 42]),
        ("TEST_BOOL", "true", True),
        ("TEST_NUMBER", "42", 42),
    ],
)
def test_substitute_env_vars(
    monkeypatch: pytest.MonkeyPatch, var_name: str, var_value: str, expected_result: Any
) -> None:
    monkeypatch.setenv(var_name, var_value)
    result = config.substitute_env_vars({"env": var_name})
    assert result == expected_result


def test_substitute_env_vars_nested_dict(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("NESTED_VALUE", "nested_result")
    input_dict = {
        "level1": {"level2": {"env_var": {"env": "NESTED_VALUE"}, "regular": "value"}},
        "other": "data",
    }

    result = config.substitute_env_vars(input_dict)
    assert result == {
        "level1": {"level2": {"env_var": "nested_result", "regular": "value"}},
        "other": "data",
    }


def test_substitute_env_vars_list_with_env_vars(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setenv("LIST_ITEM", "from_env")
    input_list = [
        "regular_item",
        {"env": "LIST_ITEM"},
        {"nested": {"env": "LIST_ITEM"}},
    ]

    result = config.substitute_env_vars(input_list)
    assert result == ["regular_item", "from_env", {"nested": "from_env"}]


def test_substitute_env_vars_env_var_not_set() -> None:
    var_name = "NONEXISTENT_VAR"
    with pytest.raises(
        ValueError, match=f"environment variable '{var_name}' is not set"
    ):
        config.substitute_env_vars({"env": var_name})


def test_substitute_env_vars_dict_with_other_keys(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setenv("TEST_VAR", "test_value")
    input_dict = {"env": "TEST_VAR", "other": "key"}

    result = config.substitute_env_vars(input_dict)
    assert result == {"env": "TEST_VAR", "other": "key"}


def test_substitute_env_vars_no_changes_needed() -> None:
    input_dict = {
        "tables": [{"name": "test", "uri": "path/to/table", "format": "delta"}],
        "queries": [],
    }

    result = config.substitute_env_vars(input_dict)
    assert result == input_dict


def test_load_yaml_config_with_env_var_substitution(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    monkeypatch.setenv("TABLE_NAME", "env_table")
    monkeypatch.setenv("TABLE_URI", "env/path/to/table")
    monkeypatch.setenv("TABLE_FORMAT", "delta")

    config_dict = {
        "tables": [
            {
                "name": {"env": "TABLE_NAME"},
                "uri": {"env": "TABLE_URI"},
                "format": {"env": "TABLE_FORMAT"},
            }
        ],
        "queries": [],
    }

    config_path = tmp_path / "test_config.yml"
    config_path.write_text(yaml.dump(config_dict))

    conf = config.load_yaml_config(config_path)

    assert len(conf.tables) == 1
    assert conf.tables[0].name == "env_table"
    assert conf.tables[0].uri == "env/path/to/table"
    assert conf.tables[0].table_format.value == "delta"



import argparse
import os
from pathlib import Path

import rich.jupyter
import rich.panel
import rich.table
import rich.text
import rich.tree
import uvicorn
import pyarrow.csv as pacsv

from laketower.__about__ import __version__
from laketower.config import load_yaml_config
from laketower.tables import (
    ImportFileFormatEnum,
    ImportModeEnum,
    execute_query,
    extract_query_parameter_names,
    generate_table_query,
    generate_table_statistics_query,
    import_file_to_table,
    load_datasets,
    load_table,
)


def run_web(config_path: Path, reload: bool) -> None:  # pragma: no cover
    os.environ["LAKETOWER_CONFIG_PATH"] = str(config_path.absolute())
    uvicorn.run("laketower.web:create_app", factory=True, reload=reload)


def validate_config(config_path: Path) -> None:
    console = rich.get_console()
    try:
        config = load_yaml_config(config_path)
        console.print(rich.panel.Panel.fit("[green]Configuration is valid"))
        console.print(config)
    except Exception as e:
        console.print(rich.panel.Panel.fit("[red]Configuration is invalid"))
        console.print(e)


def list_tables(config_path: Path) -> None:
    config = load_yaml_config(config_path)
    tree = rich.tree.Tree("tables")
    for table in config.tables:
        table_tree = tree.add(table.name)
        table_tree.add(f"format: {table.table_format.value}")
        table_tree.add(f"uri: {table.uri}")
    console = rich.get_console()
    console.print(tree)


def table_metadata(config_path: Path, table_name: str) -> None:
    out: rich.jupyter.JupyterMixin
    try:
        config = load_yaml_config(config_path)
        table_config = next(filter(lambda x: x.name == table_name, config.tables))
        table = load_table(table_config)
        metadata = table.metadata()

        out = rich.tree.Tree(table_name)
        out.add(f"name: {metadata.name}")
        out.add(f"description: {metadata.description}")
        out.add(f"format: {metadata.table_format.value}")
        out.add(f"uri: {metadata.uri}")
        out.add(f"id: {metadata.id}")
        out.add(f"version: {metadata.version}")
        out.add(f"created at: {metadata.created_at}")
        out.add(f"partitions: {', '.join(metadata.partitions)}")
        out.add(f"configuration: {metadata.configuration}")
    except Exception as e:
        out = rich.panel.Panel.fit(f"[red]{e}")

    console = rich.get_console()
    console.print(out)


def table_schema(config_path: Path, table_name: str) -> None:
    out: rich.jupyter.JupyterMixin
    try:
        config = load_yaml_config(config_path)
        table_config = next(filter(lambda x: x.name == table_name, config.tables))
        table = load_table(table_config)
        schema = table.schema()

        out = rich.tree.Tree(table_name)
        for field in schema:
            nullable = "" if field.nullable else " not null"
            out.add(f"{field.name}: {field.type}{nullable}")
    except Exception as e:
        out = rich.panel.Panel.fit(f"[red]{e}")

    console = rich.get_console()
    console.print(out, markup=False)  # disable markup to allow bracket characters


def table_history(config_path: Path, table_name: str) -> None:
    out: rich.jupyter.JupyterMixin
    try:
        config = load_yaml_config(config_path)
        table_config = next(filter(lambda x: x.name == table_name, config.tables))
        table = load_table(table_config)
        history = table.history()

        out = rich.tree.Tree(table_name)
        for rev in history.revisions:
            tree_version = out.add(f"version: {rev.version}")
            tree_version.add(f"timestamp: {rev.timestamp}")
            tree_version.add(f"client version: {rev.client_version}")
            tree_version.add(f"operation: {rev.operation}")
            tree_op_params = tree_version.add("operation parameters")
            for param_key, param_val in rev.operation_parameters.items():
                tree_op_params.add(f"{param_key}: {param_val}")
            tree_op_metrics = tree_version.add("operation metrics")
            for metric_key, metric_val in rev.operation_metrics.items():
                tree_op_metrics.add(f"{metric_key}: {metric_val}")
    except Exception as e:
        out = rich.panel.Panel.fit(f"[red]{e}")

    console = rich.get_console()
    console.print(out, markup=False)


def table_statistics(
    config_path: Path, table_name: str, version: int | None = None
) -> None:
    out: rich.jupyter.JupyterMixin
    try:
        config = load_yaml_config(config_path)
        table_config = next(filter(lambda x: x.name == table_name, config.tables))
        table = load_table(table_config)
        table_dataset = table.dataset(version=version)
        sql_query = generate_table_statistics_query(table_name)
        results = execute_query({table_name: table_dataset}, sql_query)

        out = rich.table.Table()
        for column in results.column_names:
            out.add_column(column)
        for row_dict in results.to_pylist():
            out.add_row(*[str(row_dict[col]) for col in results.column_names])
    except Exception as e:
        out = rich.panel.Panel.fit(f"[red]{e}")

    console = rich.get_console()
    console.print(out, markup=False)  # disable markup to allow bracket characters


def view_table(
    config_path: Path,
    table_name: str,
    limit: int | None = None,
    cols: list[str] | None = None,
    sort_asc: str | None = None,
    sort_desc: str | None = None,
    version: int | None = None,
) -> None:
    out: rich.jupyter.JupyterMixin
    try:
        config = load_yaml_config(config_path)
        table_config = next(filter(lambda x: x.name == table_name, config.tables))
        table = load_table(table_config)
        table_dataset = table.dataset(version=version)
        sql_query = generate_table_query(
            table_name, limit=limit, cols=cols, sort_asc=sort_asc, sort_desc=sort_desc
        )
        results = execute_query({table_name: table_dataset}, sql_query)

        out = rich.table.Table()
        for column in results.column_names:
            out.add_column(column)
        for row_dict in results.to_pylist():
            out.add_row(*[str(row_dict[col]) for col in results.column_names])
    except Exception as e:
        out = rich.panel.Panel.fit(f"[red]{e}")

    console = rich.get_console()
    console.print(out)


def query_table(
    config_path: Path,
    sql_query: str,
    sql_params: list[list[str]] = [],
    output_path: Path | None = None,
) -> None:
    out: rich.jupyter.JupyterMixin
    try:
        config = load_yaml_config(config_path)
        tables_dataset = load_datasets(config.tables)
        sql_params_dict = {param[0]: param[1] for param in sql_params}
        query_param_names = extract_query_parameter_names(sql_query)
        query_params = {
            name: sql_params_dict.get(name) or "" for name in query_param_names
        }
        results = execute_query(tables_dataset, sql_query, sql_params=query_params)

        out = rich.table.Table()
        for column in results.column_names:
            out.add_column(column)
        for row_dict in results.to_pylist():
            out.add_row(*[str(row_dict[col]) for col in results.column_names])

        if output_path is not None:
            pacsv.write_csv(
                results,
                output_path,
                pacsv.WriteOptions(include_header=True, delimiter=","),
            )
            out = rich.text.Text(f"Query results written to: {output_path}")
    except ValueError as e:
        out = rich.panel.Panel.fit(f"[red]{e}")

    console = rich.get_console()
    console.print(out)


def import_table(
    config_path: Path,
    table_name: str,
    file_path: Path,
    mode: ImportModeEnum,
    file_format: ImportFileFormatEnum,
    delimiter: str,
    encoding: str,
) -> None:
    out: rich.jupyter.JupyterMixin
    try:
        config = load_yaml_config(config_path)
        table_config = next(filter(lambda x: x.name == table_name, config.tables))
        with open(file_path, "rb") as file_content:
            rows_imported = import_file_to_table(
                table_config, file_content, mode, file_format, delimiter, encoding
            )
        out = rich.text.Text(
            f"Successfully imported {rows_imported} rows into table '{table_name}' in '{mode.value}' mode"
        )
    except Exception as e:
        out = rich.panel.Panel.fit(f"[red]{e}")

    console = rich.get_console()
    console.print(out)


def list_queries(config_path: Path) -> None:
    config = load_yaml_config(config_path)
    tree = rich.tree.Tree("queries")
    for query in config.queries:
        tree.add(query.name)
    console = rich.get_console()
    console.print(tree)


def view_query(
    config_path: Path, query_name: str, query_params: list[list[str]] = []
) -> None:
    out: rich.jupyter.JupyterMixin
    try:
        config = load_yaml_config(config_path)
        tables_dataset = load_datasets(config.tables)
        query_config = next(filter(lambda x: x.name == query_name, config.queries))
        default_parameters = {k: v.default for k, v in query_config.parameters.items()}
        sql_query = query_config.sql
        query_params_dict = {param[0]: param[1] for param in query_params}
        sql_param_names = extract_query_parameter_names(sql_query)
        sql_params = {
            name: query_params_dict.get(name) or default_parameters.get(name) or ""
            for name in sql_param_names
        }
        results = execute_query(tables_dataset, sql_query, sql_params=sql_params)

        out = rich.table.Table()
        for column in results.column_names:
            out.add_column(column)
        for row_dict in results.to_pylist():
            out.add_row(*[str(row_dict[col]) for col in results.column_names])
    except ValueError as e:
        out = rich.panel.Panel.fit(f"[red]{e}")

    console = rich.get_console()
    console.print(out)


def cli() -> None:
    parser = argparse.ArgumentParser(
        "laketower", formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument("--version", action="version", version=__version__)
    parser.add_argument(
        "--config",
        "-c",
        default="laketower.yml",
        type=Path,
        help="Path to the Laketower YAML configuration file",
    )
    subparsers = parser.add_subparsers(title="commands", required=True)

    parser_web = subparsers.add_parser(
        "web", help="Launch the web application", add_help=True
    )
    parser_web.add_argument(
        "--reload",
        help="Reload the web server on changes",
        action="store_true",
        required=False,
    )
    parser_web.set_defaults(func=lambda x: run_web(x.config, x.reload))

    parser_config = subparsers.add_parser(
        "config", help="Work with configuration", add_help=True
    )
    subsparsers_config = parser_config.add_subparsers(required=True)

    parser_config_validate = subsparsers_config.add_parser(
        "validate", help="Validate YAML configuration"
    )
    parser_config_validate.set_defaults(func=lambda x: validate_config(x.config))

    parser_tables = subparsers.add_parser("tables", help="Work with tables")
    subsparsers_tables = parser_tables.add_subparsers(required=True)

    parser_tables_list = subsparsers_tables.add_parser(
        "list", help="List all registered tables"
    )
    parser_tables_list.set_defaults(func=lambda x: list_tables(x.config))

    parser_tables_metadata = subsparsers_tables.add_parser(
        "metadata", help="Display a given table metadata"
    )
    parser_tables_metadata.add_argument("table", help="Name of the table")
    parser_tables_metadata.set_defaults(
        func=lambda x: table_metadata(x.config, x.table)
    )

    parser_tables_schema = subsparsers_tables.add_parser(
        "schema", help="Display a given table schema"
    )
    parser_tables_schema.add_argument("table", help="Name of the table")
    parser_tables_schema.set_defaults(func=lambda x: table_schema(x.config, x.table))

    parser_tables_history = subsparsers_tables.add_parser(
        "history", help="Display the history of a given table schema"
    )
    parser_tables_history.add_argument("table", help="Name of the table")
    parser_tables_history.set_defaults(func=lambda x: table_history(x.config, x.table))

    parser_tables_statistics = subsparsers_tables.add_parser(
        "statistics", help="Display summary statistics of a given table schema"
    )
    parser_tables_statistics.add_argument("table", help="Name of the table")
    parser_tables_statistics.add_argument(
        "--version", type=int, help="Time-travel to table revision number"
    )
    parser_tables_statistics.set_defaults(
        func=lambda x: table_statistics(x.config, x.table, x.version)
    )

    parser_tables_view = subsparsers_tables.add_parser(
        "view", help="View a given table"
    )
    parser_tables_view.add_argument("table", help="Name of the table")
    parser_tables_view.add_argument(
        "--limit", type=int, help="Maximum number of rows to display"
    )
    parser_tables_view.add_argument("--cols", nargs="*", help="Columns to display")
    parser_tables_view_sort_group = parser_tables_view.add_mutually_exclusive_group()
    parser_tables_view_sort_group.add_argument(
        "--sort-asc", help="Sort by given column in ascending order"
    )
    parser_tables_view_sort_group.add_argument(
        "--sort-desc", help="Sort by given column in descending order"
    )
    parser_tables_view.add_argument(
        "--version", type=int, help="Time-travel to table revision number"
    )
    parser_tables_view.set_defaults(
        func=lambda x: view_table(
            x.config, x.table, x.limit, x.cols, x.sort_asc, x.sort_desc, x.version
        )
    )

    parser_tables_query = subsparsers_tables.add_parser(
        "query", help="Query registered tables"
    )
    parser_tables_query.add_argument(
        "--output", help="Output query results to a file (default format: CSV)"
    )
    parser_tables_query.add_argument(
        "--param",
        "-p",
        nargs=2,
        action="append",
        default=[],
        help="Inject query named parameters values",
    )
    parser_tables_query.add_argument("sql", help="SQL query to execute")
    parser_tables_query.set_defaults(
        func=lambda x: query_table(x.config, x.sql, x.param, x.output)
    )

    parser_tables_import = subsparsers_tables.add_parser(
        "import", help="Import data into a table"
    )
    parser_tables_import.add_argument("table", help="Name of the table")
    parser_tables_import.add_argument(
        "--file", type=Path, required=True, help="Path to file to import"
    )
    parser_tables_import.add_argument(
        "--mode",
        choices=[mode.value for mode in ImportModeEnum],
        default=ImportModeEnum.append.value,
        type=ImportModeEnum,
        help=f"Import mode (default: {ImportModeEnum.append.value})",
    )
    parser_tables_import.add_argument(
        "--format",
        choices=[file_format.value for file_format in ImportFileFormatEnum],
        default=ImportFileFormatEnum.csv.value,
        type=ImportFileFormatEnum,
        help=f"File format (default: {ImportFileFormatEnum.csv.value})",
    )
    parser_tables_import.add_argument(
        "--delimiter", default=",", help="Column delimiter to use (default: ',')"
    )
    parser_tables_import.add_argument(
        "--encoding", default="utf-8", help="File encoding to use (default: 'utf-8')"
    )
    parser_tables_import.set_defaults(
        func=lambda x: import_table(
            x.config, x.table, x.file, x.mode, x.format, x.delimiter, x.encoding
        )
    )

    parser_queries = subparsers.add_parser("queries", help="Work with queries")
    subsparsers_queries = parser_queries.add_subparsers(required=True)

    parser_queries_list = subsparsers_queries.add_parser(
        "list", help="List all registered queries"
    )
    parser_queries_list.set_defaults(func=lambda x: list_queries(x.config))

    parser_queries_view = subsparsers_queries.add_parser(
        "view", help="View a given query"
    )
    parser_queries_view.add_argument("query", help="Name of the query")
    parser_queries_view.add_argument(
        "--param",
        "-p",
        nargs=2,
        action="append",
        default=[],
        help="Inject query named parameters values",
    )
    parser_queries_view.set_defaults(
        func=lambda x: view_query(x.config, x.query, x.param)
    )

    args = parser.parse_args()
    args.func(args)



import io
import urllib.parse
from dataclasses import dataclass
from pathlib import Path
from typing import Annotated

import bleach
import markdown
import pyarrow.csv as pacsv
import pydantic_settings
from fastapi import APIRouter, FastAPI, File, Form, Query, Request, UploadFile
from fastapi.responses import HTMLResponse, RedirectResponse, Response
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates

from laketower import __about__
from laketower.config import Config, load_yaml_config
from laketower.tables import (
    DEFAULT_LIMIT,
    ImportFileFormatEnum,
    ImportModeEnum,
    execute_query,
    extract_query_parameter_names,
    generate_table_statistics_query,
    generate_table_query,
    import_file_to_table,
    load_datasets,
    load_table,
)


class Settings(pydantic_settings.BaseSettings):
    laketower_config_path: Path


@dataclass(frozen=True)
class AppMetadata:
    app_name: str
    app_version: str


def current_path_with_args(request: Request, args: list[tuple[str, str]]) -> str:
    keys_to_update = set(arg[0] for arg in args)
    query_params = request.query_params.multi_items()
    new_query_params = list(
        filter(lambda param: param[0] not in keys_to_update, query_params)
    )
    new_query_params.extend((k, v) for k, v in args if v is not None)
    query_string = urllib.parse.urlencode(new_query_params)
    return f"{request.url.path}?{query_string}"


def render_markdown(md_text: str) -> str:
    return bleach.clean(
        markdown.markdown(md_text), tags=bleach.sanitizer.ALLOWED_TAGS | {"p"}
    )


templates = Jinja2Templates(directory=Path(__file__).parent / "templates")
templates.env.filters["current_path_with_args"] = current_path_with_args
templates.env.filters["render_markdown"] = render_markdown

router = APIRouter()


@router.get("/", response_class=HTMLResponse)
def index(request: Request) -> HTMLResponse:
    app_metadata: AppMetadata = request.app.state.app_metadata
    config: Config = request.app.state.config
    return templates.TemplateResponse(
        request=request,
        name="index.html",
        context={
            "app_metadata": app_metadata,
            "tables": config.tables,
            "queries": config.queries,
        },
    )


@router.get("/tables/query", response_class=HTMLResponse)
def get_tables_query(request: Request, sql: str) -> HTMLResponse:
    app_metadata: AppMetadata = request.app.state.app_metadata
    config: Config = request.app.state.config
    tables_dataset = load_datasets(config.tables)
    sql_schema = {
        table_name: dataset.schema.names
        for table_name, dataset in tables_dataset.items()
    }
    sql_param_names = extract_query_parameter_names(sql)
    sql_params = {
        name: request.query_params.get(name) or "" for name in sql_param_names
    }

    try:
        results = execute_query(tables_dataset, sql, sql_params=sql_params)
        error = None
    except ValueError as e:
        error = {"message": str(e)}
        results = None

    return templates.TemplateResponse(
        request=request,
        name="tables/query.html",
        context={
            "app_metadata": app_metadata,
            "tables": config.tables,
            "queries": config.queries,
            "table_results": results,
            "sql_query": sql,
            "sql_schema": sql_schema,
            "sql_params": sql_params,
            "error": error,
        },
    )


@router.get("/tables/query/csv")
def export_tables_query_csv(request: Request, sql: str) -> Response:
    config: Config = request.app.state.config
    tables_dataset = load_datasets(config.tables)

    results = execute_query(tables_dataset, sql)
    csv_content = io.BytesIO()
    pacsv.write_csv(
        results, csv_content, pacsv.WriteOptions(include_header=True, delimiter=",")
    )

    return Response(
        content=csv_content.getvalue(),
        media_type="text/csv",
        headers={"Content-Disposition": "attachment; filename=query_results.csv"},
    )


@router.get("/tables/{table_id}", response_class=HTMLResponse)
def get_table_index(request: Request, table_id: str) -> HTMLResponse:
    app_metadata: AppMetadata = request.app.state.app_metadata
    config: Config = request.app.state.config
    table_config = next(
        filter(lambda table_config: table_config.name == table_id, config.tables)
    )
    try:
        table = load_table(table_config)
        table_metadata = table.metadata()
        table_schema = table.schema()
        error = None
    except ValueError as e:
        error = {"message": str(e)}
        table_metadata = None
        table_schema = None

    return templates.TemplateResponse(
        request=request,
        name="tables/index.html",
        context={
            "app_metadata": app_metadata,
            "tables": config.tables,
            "queries": config.queries,
            "table_id": table_id,
            "table_metadata": table_metadata,
            "table_schema": table_schema,
            "error": error,
        },
    )


@router.get("/tables/{table_id}/history", response_class=HTMLResponse)
def get_table_history(request: Request, table_id: str) -> HTMLResponse:
    app_metadata: AppMetadata = request.app.state.app_metadata
    config: Config = request.app.state.config
    table_config = next(
        filter(lambda table_config: table_config.name == table_id, config.tables)
    )
    try:
        table = load_table(table_config)
        table_history = table.history()
        error = None
    except ValueError as e:
        error = {"message": str(e)}
        table_history = None

    return templates.TemplateResponse(
        request=request,
        name="tables/history.html",
        context={
            "app_metadata": app_metadata,
            "tables": config.tables,
            "queries": config.queries,
            "table_id": table_id,
            "table_history": table_history,
            "error": error,
        },
    )


@router.get("/tables/{table_id}/statistics", response_class=HTMLResponse)
def get_table_statistics(
    request: Request,
    table_id: str,
    version: int | None = None,
) -> HTMLResponse:
    app_metadata: AppMetadata = request.app.state.app_metadata
    config: Config = request.app.state.config
    table_config = next(
        filter(lambda table_config: table_config.name == table_id, config.tables)
    )
    try:
        table = load_table(table_config)
        table_name = table_config.name
        table_metadata = table.metadata()
        table_dataset = table.dataset(version=version)
        sql_query = generate_table_statistics_query(table_name)
        query_results = execute_query({table_name: table_dataset}, sql_query)
        error = None
    except ValueError as e:
        error = {"message": str(e)}
        table_metadata = None
        query_results = None

    return templates.TemplateResponse(
        request=request,
        name="tables/statistics.html",
        context={
            "app_metadata": app_metadata,
            "tables": config.tables,
            "queries": config.queries,
            "table_id": table_id,
            "table_metadata": table_metadata,
            "table_results": query_results,
            "error": error,
        },
    )


@router.get("/tables/{table_id}/view", response_class=HTMLResponse)
def get_table_view(
    request: Request,
    table_id: str,
    limit: int | None = None,
    cols: Annotated[list[str] | None, Query()] = None,
    sort_asc: str | None = None,
    sort_desc: str | None = None,
    version: int | None = None,
) -> HTMLResponse:
    app_metadata: AppMetadata = request.app.state.app_metadata
    config: Config = request.app.state.config
    table_config = next(
        filter(lambda table_config: table_config.name == table_id, config.tables)
    )
    try:
        table = load_table(table_config)
        table_name = table_config.name
        table_metadata = table.metadata()
        table_dataset = table.dataset(version=version)
        sql_query = generate_table_query(
            table_name, limit=limit, cols=cols, sort_asc=sort_asc, sort_desc=sort_desc
        )
        results = execute_query({table_name: table_dataset}, sql_query)
        error = None
    except ValueError as e:
        error = {"message": str(e)}
        table_metadata = None
        sql_query = None
        results = None

    return templates.TemplateResponse(
        request=request,
        name="tables/view.html",
        context={
            "app_metadata": app_metadata,
            "tables": config.tables,
            "queries": config.queries,
            "table_id": table_id,
            "table_metadata": table_metadata,
            "table_results": results,
            "sql_query": sql_query,
            "default_limit": DEFAULT_LIMIT,
            "error": error,
        },
    )


@router.get("/tables/{table_id}/import", response_class=HTMLResponse)
def get_table_import(
    request: Request,
    table_id: str,
) -> HTMLResponse:
    app_metadata: AppMetadata = request.app.state.app_metadata
    config: Config = request.app.state.config
    table_config = next(
        filter(lambda table_config: table_config.name == table_id, config.tables)
    )
    try:
        table = load_table(table_config)
        table_metadata = table.metadata()
        message = None
    except ValueError as e:
        message = {"type": "error", "body": str(e)}
        table_metadata = None

    return templates.TemplateResponse(
        request=request,
        name="tables/import.html",
        context={
            "app_metadata": app_metadata,
            "tables": config.tables,
            "queries": config.queries,
            "table_id": table_id,
            "table_metadata": table_metadata,
            "message": message,
        },
    )


@router.post("/tables/{table_id}/import", response_class=HTMLResponse)
def post_table_import(
    request: Request,
    table_id: str,
    input_file: Annotated[UploadFile, File()],
    mode: Annotated[ImportModeEnum, Form()],
    file_format: Annotated[ImportFileFormatEnum, Form()],
    delimiter: Annotated[str, Form()],
    encoding: Annotated[str, Form()],
) -> HTMLResponse:
    app_metadata: AppMetadata = request.app.state.app_metadata
    config: Config = request.app.state.config
    table_config = next(
        filter(lambda table_config: table_config.name == table_id, config.tables)
    )
    try:
        table = load_table(table_config)
        table_metadata = table.metadata()
        rows_imported = import_file_to_table(
            table_config, input_file.file, mode, file_format, delimiter, encoding
        )
        message = {
            "type": "success",
            "body": f"Successfully imported {rows_imported} rows",
        }
    except Exception as e:
        message = {"type": "error", "body": str(e)}
        table_metadata = None

    return templates.TemplateResponse(
        request=request,
        name="tables/import.html",
        context={
            "app_metadata": app_metadata,
            "tables": config.tables,
            "queries": config.queries,
            "table_id": table_id,
            "table_metadata": table_metadata,
            "message": message,
        },
    )


@router.get("/queries/{query_id}/view", response_class=HTMLResponse)
def get_query_view(request: Request, query_id: str) -> Response:
    app_metadata: AppMetadata = request.app.state.app_metadata
    config: Config = request.app.state.config
    query_config = next(
        filter(lambda query_config: query_config.name == query_id, config.queries)
    )

    if (
        len(request.query_params.keys()) == 0
        and len(query_config.parameters.keys()) > 0
    ):
        default_parameters = {k: v.default for k, v in query_config.parameters.items()}
        url = request.url_for("get_query_view", query_id=query_id)
        query_params = urllib.parse.urlencode(default_parameters)
        return RedirectResponse(f"{url}?{query_params}")

    tables_dataset = load_datasets(config.tables)
    sql_param_names = extract_query_parameter_names(query_config.sql)
    sql_params = {
        name: request.query_params.get(name)
        or (
            query_param.default
            if (query_param := query_config.parameters.get(name))
            else None
        )
        or ""
        for name in sql_param_names
    }

    try:
        results = execute_query(tables_dataset, query_config.sql, sql_params=sql_params)
        error = None
    except ValueError as e:
        error = {"message": str(e)}
        results = None

    return templates.TemplateResponse(
        request=request,
        name="queries/view.html",
        context={
            "app_metadata": app_metadata,
            "tables": config.tables,
            "queries": config.queries,
            "query": query_config,
            "query_results": results,
            "sql_params": sql_params,
            "error": error,
        },
    )


def create_app() -> FastAPI:
    settings = Settings()  # type: ignore[call-arg]
    config = load_yaml_config(settings.laketower_config_path)

    app = FastAPI(title="laketower")
    app.mount(
        "/static",
        StaticFiles(directory=Path(__file__).parent / "static"),
        name="static",
    )
    app.include_router(router)
    app.state.app_metadata = AppMetadata(
        app_name="🗼 Laketower", app_version=__about__.__version__
    )
    app.state.config = config

    return app



import enum
import json
import os
from pathlib import Path
from typing import Any

import pydantic
import yaml


def substitute_env_vars(config_data: Any) -> Any:
    """
    Substitute environment variables within the input payload.

    Only allowed format:
    ```python
    {
        "some_key": {"env": "VAR_NAME"}
    }

    If the "env" key MUST BE the only key in the dict to be processed.

    The content of the environment variable will be loaded with a JSON parser,
    so it can contain complex and nested structures (default is a string).
    ```
    """
    match config_data:
        case {"env": str(var_name)} if len(config_data) == 1:
            # Handle environment variable substitution
            env_value = os.getenv(var_name)
            if env_value is None:
                raise ValueError(f"environment variable '{var_name}' is not set")

            try:
                return json.loads(env_value)
            except json.JSONDecodeError:
                return env_value

        case dict() as config_dict:
            # Process dictionary recursively
            return {
                key: substitute_env_vars(value) for key, value in config_dict.items()
            }

        case list() as config_list:
            # Process list recursively
            return [substitute_env_vars(item) for item in config_list]

        case _:
            # Return primitive values unchanged
            return config_data


class TableFormats(str, enum.Enum):
    delta = "delta"


class ConfigTableConnectionS3(pydantic.BaseModel):
    s3_access_key_id: str
    s3_secret_access_key: pydantic.SecretStr
    s3_region: str | None = None
    s3_endpoint_url: pydantic.AnyHttpUrl | None = None
    s3_allow_http: bool = False


class ConfigTableConnectionADLS(pydantic.BaseModel):
    adls_account_name: str
    adls_access_key: pydantic.SecretStr | None = None
    adls_sas_key: pydantic.SecretStr | None = None
    adls_tenant_id: str | None = None
    adls_client_id: str | None = None
    adls_client_secret: pydantic.SecretStr | None = None
    azure_msi_endpoint: pydantic.AnyHttpUrl | None = None
    use_azure_cli: bool = False


class ConfigTableConnection(pydantic.BaseModel):
    s3: ConfigTableConnectionS3 | None = None
    adls: ConfigTableConnectionADLS | None = None

    @pydantic.model_validator(mode="after")
    def mutually_exclusive_connectors(self) -> "ConfigTableConnection":
        connectors = [self.s3, self.adls]
        non_null_connectors = list(filter(None, connectors))
        if len(non_null_connectors) > 1:
            raise ValueError(
                "only one connection type can be specified among: 's3', 'adls'"
            )
        return self


class ConfigTable(pydantic.BaseModel):
    name: str
    uri: str
    table_format: TableFormats = pydantic.Field(alias="format")
    connection: ConfigTableConnection | None = None


class ConfigQueryParameter(pydantic.BaseModel):
    default: str


class ConfigQuery(pydantic.BaseModel):
    name: str
    title: str
    description: str | None = None
    parameters: dict[str, ConfigQueryParameter] = {}
    sql: str


class Config(pydantic.BaseModel):
    tables: list[ConfigTable] = []
    queries: list[ConfigQuery] = []


def load_yaml_config(config_path: Path) -> Config:
    config_dict = yaml.safe_load(config_path.read_text())
    config_dict = substitute_env_vars(config_dict)
    return Config.model_validate(config_dict)
